# w Oό`@\  Ver.1.24

ConfigurationFile = "heron.cfg"
require 'win32ole'

# CLASS: ϕ\NX
class HeronsTable
  attr  :configFile,	true
  attr  :path,				true
	attr	:subtotal,		true
	attr	:viaHeron,		true

  def initialize
    @configFile = ConfigurationFile
    @path = ''
    @subtotal = false
    @viaHeron = true
  end

  def readConfig
    if @configFile[ 1, 1 ] != ":"
      config_file = @path + @configFile
    else
      config_file = @configFile
    end

    file = open( config_file, 'r' )
    begin
      @titles = Array.new
      @columns = Array.new
      @footers = Array.new
      @borders = Array.new
      @loopBorder = HeronsBorder.new( "" )
      @commands = Array.new

      while file.gets
        case $_
          when /^[ |\t]*(BEGIN|LOOP|END|LINE|WIDE|VSEP|ULINE|SET|COMMA)/
            action = $1
          when /^FIN/
            break
          when /^#/
            next
          else
            $_.chomp!
            case action
              when /^BEGIN$/
                @titles.push $_
              when /^LOOP$/
                @columns.push $_
              when /^END$/
                @footers.push $_
              when /^LINE$/
                @borders.push HeronsBorder.new( $_ )
              when /^WIDE$/
                @columnWidth = split( "," )
                @columnWidth.collect!{ |item| item.to_f }
              when /^VSEP$/
                @vsep = $_.to_f
              when /^ULINE$/
                @loopBorder.set $_
              when /^SET/
                @commands.push $_
              when /^COMMA$/
                @comma = $_
            end
        end
      end
    ensure
      file.close
    end
  end

  # METHOD: \̏
  def draw( triangle_groups, origin, font, formats, out )
    x = Array.new
    y = Array.new

    drawCommands( font, out )

    x.push origin.x
    for width in @columnWidth
      x.push ( x.last + width * ( font.width + font.gap ) )
    end

    y.push origin.y
    drawTitle( x, y, font, out )
		@titleRow		= y.length - 1
		@loopStart	= @titleRow + 1
		@loopEnd		= @loopStart + 1
		@totalRow		= @loopEnd + 1
		@areaRow		= @totalRow + 1

    y.push y.last
    total = drawLoop( triangle_groups, x, y, font, formats, out )
    drawFooter( total, x, y, font, formats, out )
    drawBorders( x, y, out )
  end

  # METHOD: R}h̏ (PRIVATE)
  private
  def drawCommands( font, out )
    out.print "lc1\n"   # Default Pen Color
    for com in @commands
      case com
        when /^cn/
          font.id = com[ 2, 1 ].to_i
        when /^lg/
          scale.layerGroup = com[ 2, 1 ].hex
      end

      out.printf "%s\n", com
    end
  end

  # METHOD: \̏ (PRIVATE)
  private
  def drawTitle( x, y, font, out )
    for title in @titles
      column_data = title.split(",")
      y.push drawRow( column_data, x, y.last, font, out)
    end
  end

  # METHOD: lf[^̏ (PRIVATE)
  private
  def drawLoop( triangle_groups, x, y, font, formats, out )
    @loopBorder.drawPen( out )
    column_contents = @columns[0].split(",")
    column_contents.collect!{ |item| item.strip! }
    column_data = Array.new
    total = 0.0
		subtotal_is_out = false

		for group in triangle_groups
			sub_total = 0.0
			if subtotal_is_out
				y.push y.last
				subtotal_is_out = false
			end
	    for triangle in group
	      column_data.clear
	      for column in column_contents
	        case column
	          when "%L1"
	            formats[ "LENGTH" ].number = triangle.l1.length / 1000
	            column_data.push ( formats[ "LENGTH" ].to_s )
	          when /^%L2$/
	            formats[ "LENGTH" ].number = triangle.l2.length / 1000
	            column_data.push ( formats[ "LENGTH" ].to_s )
	          when /^%L3$/
	            formats[ "LENGTH" ].number = triangle.l3.length / 1000
	            column_data.push ( formats[ "LENGTH" ].to_s )
	          when /^%BL$/
	            formats[ "LENGTH" ].number = triangle.bottomPath.length / 1000
	            column_data.push ( formats[ "LENGTH" ].to_s )
	          when /^%HL$/
	            formats[ "LENGTH" ].number = triangle.heightPath.length / 1000
	            column_data.push ( formats[ "LENGTH" ].to_s )
	          when /^%S$/
	          	if @viaHeron
	            	formats[ "AREA" ].number = triangle.heronsArea( formats[ "LENGTH" ] )
	            else
	            	formats[ "AREA" ].number = triangle.area( formats[ "LENGTH" ] )
	            end
	            column_data.push ( formats[ "AREA" ].to_s )
	          when /^%NO$/
	            column_data.push ( triangle.name )
	          when /^%HS$/
	            formats[ "S" ].number = triangle.heronsParam( formats[ "LENGTH" ] )
	            column_data.push ( formats[ "S" ].to_s )
	          when /^%DS$/
	            formats[ "AREA" ].number = triangle.paralleloArea( formats[ "LENGTH" ] )
	            column_data.push ( formats[ "AREA" ].to_s )
	          else
	            column_data.push column
	        end
	      end
				if @viaHeron
	      	area = triangle.heronsArea( formats[ "LENGTH" ] )
	      else
	      	area = triangle.area( formats[ "LENGTH" ] )
	      end
	      formats[ "AREA" ].number = area
	      area = formats[ "AREA" ].to_s.to_f
	      total += area
	      sub_total += area
	
	      y[ y.length - 1 ] = drawRow( column_data, x, y.last, font, out )
	      @loopBorder.draw( x, y, out )
	    end
	    if @subtotal
	    	drawSubtotal( sub_total, x, y, font, formats, out )
	    	@loopBorder.row_start += 1
	    	@loopBorder.row_end = @loopBorder.row_start
	    	@loopBorder.draw( x, y, out )
	    	subtotal_is_out = true
	    	@loopBorder.row_start += 1
	    	@loopBorder.row_end = @loopBorder.row_start
	    end
		end

    return total
  end

  # METHOD: vȂǂ̏ (PRIVATE)
  private
  def drawFooter( total, x, y, font, formats, out )
    column_data = Array.new
    for footer in @footers
      column_contents = footer.split( "," )
      column_contents.collect!{ |item| item.strip! }
      column_data.clear

      for column in column_contents
        case column
          when /%T1/
            formats[ "AREA" ].number = total
            column_data.push column.gsub( "%T1", formats[ "AREA" ].to_s )
          when /%T2/
            formats[ "TOTAL" ].number = total
            formats[ "AREA" ].number = total
            dummy_str1 = formats[ "AREA" ].to_s
            total_str = column.gsub( "%T2", formats[ "TOTAL" ].to_s )
            dummy_str2 = total_str.clone
            dummy_str2.gsub!( /\^[ucd]/, '' )
            if dummy_str1.length > dummy_str2.length
            	for i in 1 .. (dummy_str1.length - dummy_str2.length)
            		total_str << ' '
            	end
            end
            column_data.push total_str
          else
            column_data.push column
        end
      end
      y.push drawRow( column_data, x, y.last, font, out )
    end
  end

  # METHOD: v̏ (PRIVATE)
  def drawSubtotal( subtotal, x, y, font, formats, out )
    column_data = Array.new
    checked = false
    for footer in @footers
      column_contents = footer.split( "," )
      column_contents.collect!{ |item| item.strip! }
      column_data.clear

      for column in column_contents
        case column
          when /%T1/
            formats[ "AREA" ].number = subtotal
            column_data.push column.gsub( "%T1", formats[ "AREA" ].to_s )
          when //
          	column_data.push column.gsub( "", "" )
          	checked = true
          else
            column_data.push column
        end
      end
      if checked
      	y.push drawRow( column_data, x, y.last, font, out )
      	break
      end
    end
  end

  # METHOD: ȑ (PRIVATE)
  private
  def drawBorders( x, y, out )
    type = 0
    color = 0
    for border in @borders
    	if border.row_start == @totalRow
    		border.row_start = y.length - 1
    	end
    	if border.row_end == @totalRow
    		border.row_end = y.length - 1
    	end
    	if border.row_start == @areaRow
    		border.row_start = y.length
    	end
    	if border.row_end == @areaRow
    		border.row_end = y.length
    	end
      type = border.drawWithPen( x, y, type, color, out )

			if y.length > @areaRow
				if border.row_start <= @loopStart and border.row_end == @loopEnd
					row = @loopEnd + 1
					while row < y.length - 2
						border.row_start = row
						border.row_end = row + 1
						type = border.drawWithPen( x, y, type, color, out )
						row += 2
					end
				end
			end
    end
  end

  # METHOD: s (PRIVATE)
  private
  def drawRow( column_data, x, y, font, out )
    fy = y - ( @vsep * font.scale.value + font.height )
    for i in 0 .. @columnWidth.length - 1
      if column_data[ i ] == "" or column_data[ i ] == nil
        next
      end
      dummy_str = column_data[i].gsub( /\^[ucd]/, '' )		# Cancel sequence character
      fx = x[i] + (@columnWidth[i] * 2 - dummy_str.length) * 0.25 * (font.width + font.gap)

      if (@columnWidth[i] * 2) < dummy_str.length				# Try cut off left space if data is too long (2003/06/05)
				column_data[i].sub!(/^ +/, "")
				dummy_str = column_data[i].gsub( /\^[ucd]/, '' )		# Cancel sequence character
      	fx = x[i+1] - (dummy_str.length + 1)* 0.5 * (font.width + font.gap)		# Right Aligned
      end

      out.printf "ch %e %e %e %e \"%s\n", fx, fy, 1, 0, column_data[i]
    end
    fy -= @vsep * font.scale.value
    return fy
  end

end

# CLASS: r
class HeronsBorder
	attr	:column_start,	true
	attr	:column_end,		true
	attr	:row_start,			true
	attr	:row_end,				true
	attr	:insideLoop,		true

  def initialize( definitions )
    set( definitions )
    @insideLoop = false
  end

  def set( definitions )
    words = definitions.split
    if 8 <= words.length and words.length <= 10
      @column_start = words[ 0 ].to_i
      @column_end = words[ 2 ].to_i
      @row_start = words[ 1 ].to_i
      @row_end = words[ 3 ].to_i

      @x_offset_start = words[ 4 ].to_f
      @x_offset_end = words[ 6 ].to_f
      @y_offset_start = words[ 5 ].to_f
      @y_offset_end = words[ 7 ].to_f

      if words.length > 8
        @lineType = words[ 8 ].to_i
      end

      if words.length > 9
        @lineColor = words[ 9 ].to_i
      end
    end
  end

  def draw( x, y, out )
    if @column_start != nil
      x1 = x[ @column_start - 1] + @x_offset_start
      x2 = x[ @column_end - 1 ] + @x_offset_end
      y1 = y[ @row_start - 1 ] + @y_offset_start
      y2 = y[ @row_end - 1 ] + @y_offset_end
  
      out.printf(" %e %e %e %e\n", x1, y1, x2, y2 )
    end
  end

  def drawWithPen( x, y, type, color, out )
    if @column_start != nil
      if type != nil and @lineType != nil
        if type != @lineType
          out.printf "lt%d\n", @lineType
        end
      end

      if color != nil and @lineColor != nil
        if type != @lineColor
          out.printf "lc%d\n", @lineColor
        end
      end

      draw x, y, out
    end

    return @lineType, @lineColor
  end

  def drawPen( out )
    if @lineType != nil
      out.printf "lt%d\n", @lineType
    end

    if @lineColor != nil
      out.printf "lc%d\n", @lineColor
    end
  end

  def drawExcel( worksheet, loop_start, loop_end )
		if @insideLoop
	  	cell_start = idx_to_loop_cell( worksheet, loop_start, @column_start )
		  cell_end = idx_to_loop_cell( worksheet, loop_end, @column_end )
		else
	  	cell_start = idx_to_cell( worksheet, @row_start, @column_start, loop_start, loop_end )
		  cell_end = idx_to_cell( worksheet, @row_end, @column_end, loop_start, loop_end )
		end

    if @row_start == @row_end
      if @row_start == 1
        worksheet.range( cell_start, cell_end ).borders( WIN32OLE::XlTop ).weight = WIN32OLE::XlThin
      else
        worksheet.range( cell_start, cell_end ).borders( WIN32OLE::XlBottom ).weight = WIN32OLE::XlThin
      end

    elsif @column_start == @column_end
      if @column_start == 1
        worksheet.range( cell_start, cell_end ).borders( WIN32OLE::XlLeft ).weight = WIN32OLE::XlThin
      else
        worksheet.range( cell_start, cell_end ).borders( WIN32OLE::XlRight ).weight = WIN32OLE::XlThin
      end

    else
      direction = ( @row_start - @row_end ) * ( @column_start - @column_end )
      if direction > 0 then
        direction = WIN32OLE::XlDiagonalDown
      else
        direction = WIN32OLE::XlDiagonalUp
      end

      worksheet.range( cell_start.offset( 1, 1 ), cell_end ).borders( direction ).weight = WIN32OLE::XlThin
    end

  end

  # METHOD: rԍExcel ̃Zɕϊ
  private
  def idx_to_cell( worksheet, row, col, loop_start, loop_end )
    if row > loop_start
      new_row = row + ( loop_end - loop_start  - 1 )
    else
      new_row = row - 1
    end

    if new_row == 0 then new_row = 1 end
    new_col = col - 1
    if new_col == 0 then new_col = 1 end

		# p new_row.to_s + " - " + new_col.to_s

    return worksheet.cells( new_row, new_col )
  end

  # METHOD: rԍExcel ̃Zɕϊ
  private
  def idx_to_loop_cell( worksheet, row, col )
		new_row = row

    if new_row == 0 then new_row = 1 end
    new_col = col - 1
    if new_col == 0 then new_col = 1 end

		# p new_row.to_s + " - " + new_col.to_s

    return worksheet.cells( new_row, new_col )
  end
end

# CLASS: ϕ\GNZNX
class HeronsExcel < HeronsTable
  attr  :excelFile, true
  attr  :scale, true
  attr  :application, true

  def initialize
    super
    @scale = 1.0
  end

  # METHOD: \̏
  def draw( triangle_groups, font, formats )
    if @application == nil
      excel = WIN32OLE.new('Excel.Application') # Excel N
      WIN32OLE.const_load( excel )      # 萔[h
    else
      excel = @application
    end

    begin
      # excel.visible = true        # For Debug
      excel.displayAlerts = false
      excel.sheetsInNewWorkbook = 1     # V[g1܂܂ꂽubN쐬
      if FileTest.exist? @excelFile
        workbook = excel.workbooks.open @excelFile
        worksheet = workbook.activeSheet
        worksheet.cells.delete WIN32OLE::XlUp
        worksheet.cells.delete WIN32OLE::XlToLeft
        new_sheet = false
      else
        workbook = excel.workbooks.Add()
        worksheet = workbook.worksheets(1)
        new_sheet = true
      end

			font.scale = nil
      cell_height = mm_to_point( @vsep * 2 + font.height, false )
      font_size = mm_to_point( font.height, true )

      setColumnWidth( worksheet, font_size )

      drawTitle( worksheet, font_size, cell_height )
			@titleRow		= worksheet.usedRange.rows.count
			@loopStart	= @titleRow + 1
			@loopEnd		= @loopStart + 1
			@totalRow		= @loopEnd + 1
			@areaRow		= @totalRow + 1

      row_start, row_end = drawLoop( worksheet, triangle_groups, cell_height, formats )
      drawFooter( worksheet, row_start, row_end, cell_height, formats )
      drawBorders( worksheet, row_start, row_end )

      if new_sheet  then  workbook.saveAs @excelFile  else  workbook.save end
    ensure
      workbook.close()
      excel.quit
    end
  end

  # METHOD: Jݒ
  def setColumnWidth( worksheet, font_size )
    worksheet.cells.font.name = "lr SVbN"
    worksheet.cells.font.size = font_size

    for idx in 1 .. @columnWidth.length
      worksheet.columns( idx ).columnWidth = @columnWidth[ idx - 1 ] * 2
    end
    
  end


  # METHOD: \̏ (PRIVATE)
  def drawTitle( worksheet, font_size, cell_height )
    row = 1
    for title in @titles
      worksheet.rows(row).rowHeight = cell_height
      worksheet.rows(row).verticalAlignment = WIN32OLE::XlCenter
      column_data = title.split(",")
      for col in 1 .. column_data.length
        if ( column_data[ col - 1 ] == "" or column_data[ col - 1 ] == nil ) then next end

        cell = worksheet.cells( row, col )
        cell.value = column_data[ col - 1 ].strip!
        cell.horizontalAlignment = WIN32OLE::XlCenter
      end
      row += 1
    end
  end

  # METHOD: lf[^̏ (PRIVATE)
  private
  def drawLoop( worksheet, triangle_groups, cell_height, formats )
    column_contents = @columns[0].split(",")
    column_contents.collect!{ |item| item.strip! }
    row_start = worksheet.usedRange.rows.count + 1
    row = row_start

    cell_values = Array.new( column_contents.length, nil )
    cell_formats = Array.new( column_contents.length, nil )
    cell_formulas = Array.new( column_contents.length, nil )
    cell_alignments = Array.new( column_contents.length, nil )
    l1_column = l2_column = l3_column = s_column = nil
		hl_column = bl_column = ds_column = nil

    col = 0
    for column in column_contents
      case column
        when /(.*)%L([1-3])(.*)/
          case $2
            when "1"
              cell_values[ col ] = '%L1'
              l1_column = col
            when "2"
              cell_values[ col ] = '%L2'
              l2_column = col
            when "3"
              cell_values[ col ] = '%L3'
              l3_column = col
          end
          cell_formats[ col ] = '"' + $1 + '"' + formats[ "LENGTH" ].excel + '"' + $3 + '"'
          cell_alignments[ col ] = WIN32OLE::XlRight
        when /(.*)%(H|B)L(.*)/
        	cell_values[ col ] = '%' +$2 + 'L'
        	cell_formats[ col ] = '"' + $1 + '"' + formats[ "LENGTH" ].excel + '"' + $3 + '"'
        	cell_alignments[ col ] = WIN32OLE::XlRight
        	if $2 == "H"
        		hl_column = col
        	else
        		bl_column = col
        	end
        when /(.*)%S(.*)/
          cell_formats[ col ] = '"' + $1 + '"' + formats[ "AREA" ].excel + '"' + $2 + '"'
          cell_alignments[ col ] = WIN32OLE::XlRight
          s_column = col
        when /(.*)%DS(.*)/
          cell_formats[ col ] = '"' + $1 + '"' + formats[ "AREA" ].excel + '"' + $2 + '"'
          cell_alignments[ col ] = WIN32OLE::XlRight
          ds_column = col
        when /(.*)%HS(.*)/
          cell_formats[ col ] = '"' + $1 + '"' + formats[ "S" ].excel + '"' + $2 + '"'
          cell_alignments[ col ] = WIN32OLE::XlRight
          ss_column = col
        else
          cell_values[ col ] = column
          cell_alignments[ col ] = WIN32OLE::XlCenter
      end
      col += 1
    end

    if l1_column == nil and @viaHeron
      l1_column = cell_values.length
      worksheet.columns( l1_column + 1).columnWidth = 0.0
      cell_values.push '%L1'
      cell_formats.push formats[ "LENGTH" ].excel
      cell_alignments.push  WIN32OLE::XlRight
      cell_formulas.push nil
    end
    if l2_column == nil and @viaHeron
      l2_column = cell_values.length
      worksheet.columns( l2_column + 1).columnWidth = 0.0
      cell_values.push '%L2'
      cell_formats.push formats[ "LENGTH" ].excel
      cell_alignments.push  WIN32OLE::XlRight
      cell_formulas.push nil
    end
    if l3_column == nil and @viaHeron
      l3_column = cell_values.length
      worksheet.columns( l3_column + 1 ).columnWidth = 0.0
      cell_values.push '%L3'
      cell_formats.push formats[ "LENGTH" ].excel
      cell_alignments.push  WIN32OLE::XlRight
      cell_formulas.push nil
    end
    if ss_column == nil and @viaHeron
      ss_column = cell_values.length
      worksheet.columns( ss_column + 1 ).columnWidth = 0.0
      cell_values.push "%HS"
      cell_formats.push formats[ "S" ].excel
      cell_alignments.push  WIN32OLE::XlRight
      cell_formulas.push ''
      cell = worksheet.cells( row - 1, ss_column + 1)
      cell.value = 's'
      cell.horizontalAlignment = WIN32OLE::XlCenter
    end
    if bl_column == nil and (! @viaHeron)
      bl_column = cell_values.length
      worksheet.columns( bl_column + 1).columnWidth = 0.0
      cell_values.push '%BL'
      cell_formats.push formats[ "LENGTH" ].excel
      cell_alignments.push  WIN32OLE::XlRight
      cell_formulas.push nil
    end
    if hl_column == nil and (! @viaHeron)
      hl_column = cell_values.length
      worksheet.columns( hl_column + 1).columnWidth = 0.0
      cell_values.push '%HL'
      cell_formats.push formats[ "LENGTH" ].excel
      cell_alignments.push  WIN32OLE::XlRight
      cell_formulas.push nil
    end
    if ds_column == nil and (! @viaHeron)
      ds_column = cell_values.length
      worksheet.columns( ds_column + 1 ).columnWidth = 0.0
      cell_values.push "%DS"
      cell_formats.push formats[ "AREA" ].excel
      cell_alignments.push  WIN32OLE::XlRight
      cell_formulas.push ''
      cell = worksheet.cells( row - 1, ds_column + 1)
      cell.value = '{ʐ'
      cell.horizontalAlignment = WIN32OLE::XlCenter
    end

		if @viaHeron
	    cell_formulas[ ss_column ] = '=(RC[' + (l1_column - ss_column).to_s + 
	                                  ']+RC[' + (l2_column - ss_column).to_s + 
	                                  ']+RC[' + (l3_column - ss_column).to_s + 
	                                  '])*0.5'
		else
			cell_formulas[ ds_column ] = '=RC[' + (bl_column - ds_column).to_s + 
	                                  ']*RC[' + (hl_column - ds_column).to_s +
	                                  ']'
	  end

    if s_column != nil
    	if @viaHeron
	      ss_cell = (ss_column - s_column).to_s
	      if ! formats[ "TOTAL" ].trancated?
	        cell_formulas[ s_column ] = '=SQRT(RC[' + ss_cell +
	                                    ']*(RC[' + ss_cell + ']-RC[' + (l1_column - s_column).to_s +
	                                    '])*(RC[' + ss_cell + ']-RC[' + (l2_column - s_column).to_s +
	                                    '])*(RC[' + ss_cell + ']-RC[' + (l3_column - s_column).to_s +
	                                    ']))'
	      else
	        cell_formulas[ s_column ] = '=ROUNDDOWN(SQRT(RC[' + ss_cell +
	                                    ']*(RC[' + ss_cell + ']-RC[' + (l1_column - s_column).to_s +
	                                    '])*(RC[' + ss_cell + ']-RC[' + (l2_column - s_column).to_s +
	                                    '])*(RC[' + ss_cell + ']-RC[' + (l3_column - s_column).to_s +
	                                    '])),' + (formats["AREA"].precision - 1).to_s + ')'
	      end
			else
				ds_cell = (ds_column - s_column).to_s
				if ! formats[ "TOTAL" ].trancated?
					cell_formulas[ s_column ] = '=RC[' + ds_cell + ']*0.5'
				else
					cell_formulas[ s_column ] = '=ROUNDDOWN(RC[' + ds_cell + ']*0.5,' + (formats["AREA"].precision - 1).to_s + ')'
				end
			end
    end

		group_start = Array.new
		group_end = Array.new
    cell_columns = cell_values.length
    for group in triangle_groups
    	group_start << row
	    for triangle in group
	      worksheet.rows( row ).rowHeight = cell_height
	      worksheet.rows( row ).verticalAlignment = WIN32OLE::XlCenter
	      for col in 0 .. cell_columns - 1
	        cell = worksheet.cells( row, col + 1 )
	        if cell_values[ col ] != nil
	          case cell_values[ col ]
	            when '%L1'
	              formats[ "LENGTH" ].number = triangle.l1.length / 1000
	              cell.value = formats[ "LENGTH" ].to_s
	            when '%L2'
	              formats[ "LENGTH" ].number = triangle.l2.length / 1000
	              cell.value = formats[ "LENGTH" ].to_s
	            when '%L3'
	              formats[ "LENGTH" ].number = triangle.l3.length / 1000
	              cell.value = formats[ "LENGTH" ].to_s
	            when '%BL'
	              formats[ "LENGTH" ].number = triangle.bottomPath.length / 1000
	              cell.value = formats[ "LENGTH" ].to_s
	            when '%HL'
	              formats[ "LENGTH" ].number = triangle.heightPath.length / 1000
	              cell.value = formats[ "LENGTH" ].to_s
	            when '%NO'
	              cell.value = triangle.name
	          end
	        end
	
	        if cell_formats[ col ] != nil then    cell.numberFormatLocal = cell_formats[ col ]      end
	        if cell_alignments[ col ] != nil then cell.horizontalAlignment = cell_alignments[ col ] end
	        if cell_formulas[ col ] != nil then   cell.formulaR1C1 = cell_formulas[ col ]           end
	      end
	
	      @loopBorder.drawExcel( worksheet, row_start, row )
	      row += 1
	    end
	    if @subtotal
	    	drawSubtotal( worksheet, group_start.last, row - 1, cell_height, formats )
	    	@loopBorder.drawExcel( worksheet, row_start, row )
	    	group_end.push (row - 1)
	    	row += 1
	    end
	  end

		if group_end.length > 0
			[ group_start, group_end ]
    else
    	[ row_start, row - 1 ]
    end
  end

  # METHOD: vȂǂ̏ (PRIVATE)
  private
  def drawFooter( worksheet, loop_start, loop_end, cell_height, formats )
		if loop_start.kind_of? Array
			sum_str = ""
			for subtotal in loop_end
				if sum_str == ""
					sum_str = "R" + (subtotal + 1 ).to_s + "C"
				else
					sum_str += "+R" + (subtotal + 1 ).to_s + "C"
				end
			end
			row = loop_end.last + 2
		else
			sum_str = "SUM(R" + loop_start.to_s + "C:R" + loop_end.to_s + "C)"
			row = loop_end + 1
		end

    for footer in @footers
      worksheet.rows( row ).rowHeight = cell_height
      worksheet.rows( row ).verticalAlignment = WIN32OLE::XlCenter

      column_contents = footer.split( "," )
      column_contents.collect!{ |item| item.strip! }
      col = 1

      for column in column_contents
        cell = worksheet.cells( row, col )
        if column != nil
          column.gsub!( /m\^u2/, "u" )
        end
        case column
          when /(.*)%T1(.*)/
            cell.formulaR1C1 = '=' + sum_str
            cell.numberFormatLocal = '"' + $1 + '"' + formats[ "AREA" ].excel[0 .. -5] + '"' + $2 + '  "'
            cell.horizontalAlignment = WIN32OLE::XlRight
          when /(.*)%T2(.*)/
            if formats[ "TOTAL" ].trancated?
              cell.formulaR1C1 = '=ROUNDDOWN(' + sum_str + ',' + (formats["TOTAL"].precision - 1).to_s + ')'
            else
              cell.formulaR1C1 = '=' + sum_str
            end
            cell.numberFormatLocal = '"' + $1 + '"' + formats[ "TOTAL" ].excel[0 .. -5] + '"' + $2 + '  "'
            cell.horizontalAlignment = WIN32OLE::XlRight
          else
            cell.value = column
            cell.horizontalAlignment = WIN32OLE::XlCenter
        end
        col += 1
      end

      row += 1
    end
  end

  # METHOD: v̏ (PRIVATE)
  def drawSubtotal( worksheet, loop_start, loop_end, cell_height, formats )
    row = loop_end + 1
		checked = false

    for footer in @footers
      worksheet.rows( row ).rowHeight = cell_height
      worksheet.rows( row ).verticalAlignment = WIN32OLE::XlCenter

      column_contents = footer.split( "," )
      column_contents.collect!{ |item| item.strip! }
      col = 1

      for column in column_contents
        cell = worksheet.cells( row, col )
        if column != nil
          column.gsub!( /m\^u2/, "u" )
        end
        case column
          when /(.*)%T1(.*)/
            cell.formulaR1C1 = "=SUM(R" + loop_start.to_s + "C:R" + loop_end.to_s + "C)"
            cell.numberFormatLocal = '"' + $1 + '"' + formats[ "AREA" ].excel[0 .. -5] + '"' + $2 + '  "'
            cell.horizontalAlignment = WIN32OLE::XlRight
          when //
            cell.value = column.gsub( //, "" )
            cell.horizontalAlignment = WIN32OLE::XlCenter
            checked = true
          else
            cell.value = column
            cell.horizontalAlignment = WIN32OLE::XlCenter
        end
        col += 1
      end

      if checked
      	break
      end
      row += 1
    end
  end

  # METHOD: ȑ (PRIVATE)
  private
  def drawBorders( worksheet, loop_start, loop_end )
		if loop_start.kind_of? Array
	    for border in @borders
				if border.row_start <= @loopStart and border.row_end == @loopEnd
					for i in 0 .. loop_start.length - 1
						if i > 0
							border.insideLoop = true
						end
						border.drawExcel( worksheet, loop_start[i], loop_end[i] )
					end
				else
		    	if border.row_start >= @totalRow
	    			border.row_start += loop_start.last - 1
	    		end
	    		if border.row_end >= @totalRow
	    			border.row_end += loop_start.last - 1
	    		end
	    		border.drawExcel( worksheet, loop_start.last, loop_end.last )
	    	end

			end
		else
	    for border in @borders
	      border.drawExcel( worksheet, loop_start, loop_end )
	    end
	  end
  end

  # METHOD: s (PRIVATE)
  private
  def drawRow( worksheet, column_data, cell_height, align )
    row = worksheet.usedRange.rows.count + 1
    worksheet.rows( row ).height = cell_height
    worksheet.rows( row ).horizontalAlignment = align

    for col in 1 .. @columnWidth.length
      if column_data[ col - 1 ] == "" or column_data[ col - 1 ] == nil
        next
      end
      worksheet.cells( row, col ).value = column_data[ col - 1 ]
      
    end
  end

  # METHOD: mm GNZ̃|Cg֕ϊi1point = 0.35mmj
  private
  def mm_to_point( a, round )
    a2 = a / 0.35 * @scale
    if round
      s = sprintf("%40.0f", a2 )
      s.to_i
    else
      a2
    end
  end
end
